import contextlib
import io
import os
import sys

import pytest

import yuio.term


class TestColor:
    @pytest.mark.parametrize(
        "color,expect",
        [
            (yuio.term.Color.NONE | yuio.term.Color.NONE, yuio.term.Color.NONE),
            (yuio.term.Color.NONE | yuio.term.Color.FORE_RED, yuio.term.Color.FORE_RED),
            (yuio.term.Color.NONE | yuio.term.Color.BACK_RED, yuio.term.Color.BACK_RED),
            (
                yuio.term.Color.FORE_BLUE | yuio.term.Color.FORE_RED,
                yuio.term.Color.FORE_RED,
            ),
            (
                yuio.term.Color.BACK_BLUE | yuio.term.Color.BACK_RED,
                yuio.term.Color.BACK_RED,
            ),
            (
                yuio.term.Color.FORE_RED | yuio.term.Color.BACK_BLUE,
                yuio.term.Color(
                    fore=yuio.term.ColorValue(1), back=yuio.term.ColorValue(4)
                ),
            ),
            (
                yuio.term.Color.FORE_RED | yuio.term.Color.STYLE_BOLD,
                yuio.term.Color(fore=yuio.term.ColorValue(1), bold=True),
            ),
            (
                yuio.term.Color.STYLE_BOLD | yuio.term.Color.FORE_RED,
                yuio.term.Color(fore=yuio.term.ColorValue(1), bold=True),
            ),
            (
                yuio.term.Color.FORE_RED | yuio.term.Color.STYLE_DIM,
                yuio.term.Color(fore=yuio.term.ColorValue(1), dim=True),
            ),
            (
                yuio.term.Color.STYLE_DIM | yuio.term.Color.FORE_RED,
                yuio.term.Color(fore=yuio.term.ColorValue(1), dim=True),
            ),
            (
                yuio.term.Color.FORE_RED
                | yuio.term.Color.STYLE_BOLD
                | yuio.term.Color.STYLE_DIM,
                yuio.term.Color(fore=yuio.term.ColorValue(1), bold=True, dim=True),
            ),
            (
                yuio.term.Color.FORE_RED
                | yuio.term.Color.STYLE_BOLD
                | yuio.term.Color.STYLE_NORMAL,
                yuio.term.Color(fore=yuio.term.ColorValue(1), bold=False, dim=False),
            ),
            (
                yuio.term.Color.FORE_RED
                | yuio.term.Color.STYLE_DIM
                | yuio.term.Color.STYLE_NORMAL,
                yuio.term.Color(fore=yuio.term.ColorValue(1), bold=False, dim=False),
            ),
            (
                yuio.term.Color.FORE_RED
                | yuio.term.Color.STYLE_DIM
                | yuio.term.Color.STYLE_NORMAL
                | yuio.term.Color.STYLE_DIM,
                yuio.term.Color(fore=yuio.term.ColorValue(1), bold=False, dim=True),
            ),
            (
                yuio.term.Color.FORE_RED
                | yuio.term.Color.STYLE_DIM
                | yuio.term.Color.STYLE_NORMAL
                | yuio.term.Color.STYLE_BOLD,
                yuio.term.Color(fore=yuio.term.ColorValue(1), bold=True, dim=False),
            ),
        ],
    )
    def test_combine(self, color, expect):
        assert color == expect

    @pytest.mark.parametrize(
        "color,cap,expect",
        [
            (yuio.term.Color.NONE, yuio.term.ColorSupport.NONE, ""),
            (yuio.term.Color.NONE, yuio.term.ColorSupport.ANSI, "\x1b[m"),
            (yuio.term.Color.NONE, yuio.term.ColorSupport.ANSI_256, "\x1b[m"),
            (yuio.term.Color.NONE, yuio.term.ColorSupport.ANSI_TRUE, "\x1b[m"),
            (yuio.term.Color.FORE_RED, yuio.term.ColorSupport.NONE, ""),
            (yuio.term.Color.FORE_RED, yuio.term.ColorSupport.ANSI, "\x1b[;31m"),
            (yuio.term.Color.FORE_RED, yuio.term.ColorSupport.ANSI_256, "\x1b[;31m"),
            (yuio.term.Color.FORE_RED, yuio.term.ColorSupport.ANSI_TRUE, "\x1b[;31m"),
            (yuio.term.Color.BACK_RED, yuio.term.ColorSupport.NONE, ""),
            (yuio.term.Color.BACK_RED, yuio.term.ColorSupport.ANSI, "\x1b[;41m"),
            (yuio.term.Color.BACK_RED, yuio.term.ColorSupport.ANSI_256, "\x1b[;41m"),
            (yuio.term.Color.BACK_RED, yuio.term.ColorSupport.ANSI_TRUE, "\x1b[;41m"),
            (yuio.term.Color.fore_from_hex("#338F15"), yuio.term.ColorSupport.NONE, ""),
            (
                yuio.term.Color.fore_from_hex("#338F15"),
                yuio.term.ColorSupport.ANSI,
                "\x1b[;32m",
            ),
            (
                yuio.term.Color.fore_from_hex("#338F15"),
                yuio.term.ColorSupport.ANSI_256,
                "\x1b[;38;5;64m",
            ),
            (
                yuio.term.Color.fore_from_hex("#338F15"),
                yuio.term.ColorSupport.ANSI_TRUE,
                "\x1b[;38;2;51;143;21m",
            ),
            (yuio.term.Color.back_from_hex("#338F15"), yuio.term.ColorSupport.NONE, ""),
            (
                yuio.term.Color.back_from_hex("#338F15"),
                yuio.term.ColorSupport.ANSI,
                "\x1b[;42m",
            ),
            (
                yuio.term.Color.back_from_hex("#338F15"),
                yuio.term.ColorSupport.ANSI_256,
                "\x1b[;48;5;64m",
            ),
            (
                yuio.term.Color.back_from_hex("#338F15"),
                yuio.term.ColorSupport.ANSI_TRUE,
                "\x1b[;48;2;51;143;21m",
            ),
        ],
    )
    def test_as_code(self, color: yuio.term.Color, cap, expect):
        term = yuio.term.Term(None, None, color_support=cap)  # type: ignore
        assert color.as_code(term) == expect

    @pytest.mark.parametrize(
        "colors,coeffs,expect",
        [
            (
                [yuio.term.Color.FORE_RED, yuio.term.Color.FORE_GREEN],
                [i / 4 for i in range(5)],
                [yuio.term.Color.FORE_RED] * 5,
            ),
            (
                [yuio.term.Color.FORE_RED, yuio.term.Color.fore_from_hex("#AA0000")],
                [i / 4 for i in range(5)],
                [yuio.term.Color.FORE_RED] * 5,
            ),
            (
                [yuio.term.Color.fore_from_hex("#AA0000"), yuio.term.Color.FORE_RED],
                [i / 4 for i in range(5)],
                [yuio.term.Color.fore_from_hex("#AA0000")] * 5,
            ),
            (
                [
                    yuio.term.Color.fore_from_hex("#AA0000"),
                    yuio.term.Color.fore_from_hex("#00AA00"),
                ],
                [i / 4 for i in range(5)],
                [
                    yuio.term.Color.fore_from_hex("#AA0000"),
                    yuio.term.Color.fore_from_hex("#7F2A00"),
                    yuio.term.Color.fore_from_hex("#555500"),
                    yuio.term.Color.fore_from_hex("#2A7F00"),
                    yuio.term.Color.fore_from_hex("#00AA00"),
                ],
            ),
            (
                [
                    yuio.term.Color.fore_from_hex("#AA0000"),
                    yuio.term.Color.fore_from_hex("#00AA00"),
                    yuio.term.Color.fore_from_hex("#0000AA"),
                ],
                [i / 8 for i in range(9)],
                [
                    yuio.term.Color.fore_from_hex("#AA0000"),
                    yuio.term.Color.fore_from_hex("#7F2A00"),
                    yuio.term.Color.fore_from_hex("#555500"),
                    yuio.term.Color.fore_from_hex("#2A7F00"),
                    yuio.term.Color.fore_from_hex("#00AA00"),
                    yuio.term.Color.fore_from_hex("#007F2A"),
                    yuio.term.Color.fore_from_hex("#005555"),
                    yuio.term.Color.fore_from_hex("#002A7F"),
                    yuio.term.Color.fore_from_hex("#0000AA"),
                ],
            ),
            (
                [
                    yuio.term.Color.fore_from_hex("#AA0000"),
                    yuio.term.Color.fore_from_hex("#00AA00"),
                ],
                [0, 0.5, 1],
                [
                    yuio.term.Color.fore_from_hex("#AA0000"),
                    yuio.term.Color.fore_from_hex("#555500"),
                    yuio.term.Color.fore_from_hex("#00AA00"),
                ],
            ),
            (
                [
                    yuio.term.Color.back_from_hex("#0000AA"),
                    yuio.term.Color.back_from_hex("#00AA00"),
                ],
                [0, 0.5, 1],
                [
                    yuio.term.Color.back_from_hex("#0000AA"),
                    yuio.term.Color.back_from_hex("#005555"),
                    yuio.term.Color.back_from_hex("#00AA00"),
                ],
            ),
            (
                [
                    yuio.term.Color.fore_from_hex("#AA0000")
                    | yuio.term.Color.back_from_hex("#0000AA"),
                    yuio.term.Color.fore_from_hex("#00AA00")
                    | yuio.term.Color.back_from_hex("#00AA00"),
                ],
                [0, 0.5, 1],
                [
                    yuio.term.Color.fore_from_hex("#AA0000")
                    | yuio.term.Color.back_from_hex("#0000AA"),
                    yuio.term.Color.fore_from_hex("#555500")
                    | yuio.term.Color.back_from_hex("#005555"),
                    yuio.term.Color.fore_from_hex("#00AA00")
                    | yuio.term.Color.back_from_hex("#00AA00"),
                ],
            ),
        ],
    )
    def test_lerp(self, colors, coeffs, expect):
        lerp = yuio.term.Color.lerp(*colors)
        result = [lerp(c) for c in coeffs]
        assert result == expect

    @pytest.mark.parametrize(
        "color,expect",
        [
            (
                yuio.term.ColorValue(0).lighten(0.5),
                yuio.term.ColorValue((127, 127, 127)),
            ),
            (yuio.term.ColorValue(0).darken(0.5), yuio.term.ColorValue((0, 0, 0))),
            (
                yuio.term.ColorValue(7).lighten(0.5),
                yuio.term.ColorValue((255, 255, 255)),
            ),
            (
                yuio.term.ColorValue(7).darken(0.5),
                yuio.term.ColorValue((127, 127, 127)),
            ),
            (
                yuio.term.ColorValue.from_hex("#005555").lighten(0.5),
                yuio.term.ColorValue.from_hex("#00AAAA"),
            ),
            (
                yuio.term.ColorValue.from_hex("#005555").darken(0.5),
                yuio.term.ColorValue.from_hex("#002A2A"),
            ),
            (yuio.term.ColorValue("0").darken(0.5), yuio.term.ColorValue("0")),
            (
                yuio.term.ColorValue.from_hex("#005555").match_luminosity(
                    yuio.term.ColorValue.from_hex("#002A2A")
                ),
                yuio.term.ColorValue.from_hex("#002A2A"),
            ),
            (
                yuio.term.ColorValue.from_hex("#005555").match_luminosity(
                    yuio.term.ColorValue.from_hex("#AA4700")
                ),
                yuio.term.ColorValue.from_hex("#00AAAA"),
            ),
        ],
    )
    def test_color_operations(self, color, expect):
        assert color == expect


class TestColorizedString:
    @pytest.mark.parametrize(
        "text,width,length",
        [
            (yuio.term.ColorizedString(), 0, 0),
            (yuio.term.ColorizedString([]), 0, 0),
            (yuio.term.ColorizedString(""), 0, 0),
            (yuio.term.ColorizedString("abc"), 3, 3),
            (yuio.term.ColorizedString("абв"), 3, 3),
            (yuio.term.ColorizedString("a\nb"), 3, 3),
            (yuio.term.ColorizedString("👻"), 2, 1),
            (yuio.term.ColorizedString(["abc", "def"]), 6, 6),
            (yuio.term.ColorizedString(["abc", "👻"]), 5, 4),
            (yuio.term.ColorizedString([yuio.term.Color.FORE_RED, "abc", "def"]), 6, 6),
            (yuio.term.ColorizedString([yuio.term.Color.FORE_RED, "abc", "👻"]), 5, 4),
            (
                yuio.term.ColorizedString([yuio.term.Color.FORE_RED, "abc", "", "def"]),
                6,
                6,
            ),
        ],
    )
    def test_width_and_len(self, text, width, length):
        assert text.width == width
        assert text.len == len(text) == length
        assert bool(text) is (len(text) > 0)

    @pytest.mark.parametrize(
        "text,args,expect",
        [
            (
                [""],
                (),
                [""],
            ),
            (
                ["hello world!"],
                (),
                ["hello world!"],
            ),
            (
                ["hello %s!"],
                "username",
                ["hello username!"],
            ),
            (
                ["hello %r!"],
                "username",
                ["hello 'username'!"],
            ),
            (
                ["hello %05.2f!"],
                1.5,
                ["hello 01.50!"],
            ),
            (
                ["hello %05.2lf!"],
                1.5,
                ["hello 01.50!"],
            ),
            (
                ["hello %s %% %s!"],
                (1, 2),
                ["hello 1 % 2!"],
            ),
            (
                ["hello %s"],
                {"a": "b"},
                ["hello {'a': 'b'}"],
            ),
            (
                ["hello %s %(a)s"],
                {"a": "b"},
                ["hello {'a': 'b'} b"],
            ),
        ],
    )
    def test_format(self, text, args, expect):
        formatted = yuio.term.ColorizedString(text) % args
        assert formatted._parts == expect

    @pytest.mark.parametrize(
        "text,width,kwargs,expect",
        [
            (
                ["hello world!"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!"],
                ],
            ),
            (
                ["hello world! 15"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!", " ", "15"],
                ],
            ),
            (
                ["hello world! 👻"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!", " ", "👻"],
                ],
            ),
            (
                ["hello world! 👻👻"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!"],
                    [yuio.term.Color.NONE, "👻👻"],
                ],
            ),
            (
                ["hello world! this will wrap"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!"],
                    [yuio.term.Color.NONE, "this", " ", "will", " ", "wrap"],
                ],
            ),
            (
                ["hello world!     this will wrap"],
                15,
                {"preserve_spaces": False},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!"],
                    [yuio.term.Color.NONE, "this", " ", "will", " ", "wrap"],
                ],
            ),
            (
                ["hello     world!     this     will     wrap"],
                15,
                {"preserve_spaces": False},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!"],
                    [yuio.term.Color.NONE, "this", " ", "will", " ", "wrap"],
                ],
            ),
            (
                ["this-will-wrap-on-hyphen"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "this-", "will-", "wrap-"],
                    [yuio.term.Color.NONE, "on-", "hyphen"],
                ],
            ),
            (
                ["this.will.not.wrap.on.hyphen.this.is.too.long"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "this.will.not.w"],
                    [yuio.term.Color.NONE, "rap.on.hyphen.t"],
                    [yuio.term.Color.NONE, "his.is.too.long"],
                ],
            ),
            (
                ["newlines will\nbe\nhonored!"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "newlines", " ", "will"],
                    [yuio.term.Color.NONE, "be"],
                    [yuio.term.Color.NONE, "honored!"],
                ],
            ),
            (
                ["space before nl    \nis removed"],
                15,
                {"preserve_spaces": False},
                [
                    [yuio.term.Color.NONE, "space", " ", "before", " ", "nl"],
                    [yuio.term.Color.NONE, "is", " ", "removed"],
                ],
            ),
            (
                ["space after nl\n    is kept"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "space", " ", "after", " ", "nl"],
                    [yuio.term.Color.NONE, "    ", "is", " ", "kept"],
                ],
            ),
            (
                ["wo", "rd wo", "rd"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "wo", "rd", " ", "wo", "rd"],
                ],
            ),
            pytest.param(
                ["wo", "rd wo", "rd"],
                7,
                {},
                [
                    [yuio.term.Color.NONE, "wo", "rd"],
                    [yuio.term.Color.NONE, "wo", "rd"],
                ],
                marks=pytest.mark.xfail(
                    reason="this is a bug which I'm not sure how to fix atm"
                ),
            ),
            (
                ["hello world!"],
                15,
                {"preserve_spaces": True},
                [
                    [yuio.term.Color.NONE, "hello", " ", "world!"],
                ],
            ),
            (
                ["hello   world!"],
                15,
                {"preserve_spaces": True},
                [
                    [yuio.term.Color.NONE, "hello", "   ", "world!"],
                ],
            ),
            (
                ["hello     world!"],
                15,
                {"preserve_spaces": True},
                [
                    [yuio.term.Color.NONE, "hello", "     "],
                    [yuio.term.Color.NONE, "world!"],
                ],
            ),
            (
                ["hello                    world"],
                15,
                {"preserve_spaces": True},
                [
                    [yuio.term.Color.NONE, "hello", "          "],
                    [yuio.term.Color.NONE, "          ", "world"],
                ],
            ),
            (
                ["hello                    longlongworld"],
                15,
                {"preserve_spaces": True},
                [
                    [yuio.term.Color.NONE, "hello", "          "],
                    [yuio.term.Color.NONE, "          "],
                    [yuio.term.Color.NONE, "longlongworld"],
                ],
            ),
            (
                ["hello ", yuio.term.Color.STYLE_BOLD, "world", yuio.term.Color.NONE],
                15,
                {},
                [
                    [
                        yuio.term.Color.NONE,
                        "hello",
                        " ",
                        yuio.term.Color.STYLE_BOLD,
                        "world",
                        yuio.term.Color.NONE,
                    ]
                ],
            ),
            (
                [],
                15,
                {},
                [
                    [yuio.term.Color.NONE],
                ],
            ),
            (
                [""],
                15,
                {},
                [
                    [yuio.term.Color.NONE],
                ],
            ),
            (
                ["12345678901234 12345"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "12345678901234"],
                    [yuio.term.Color.NONE, "12345"],
                ],
            ),
            (
                ["Hello line break", yuio.term.Color.NONE],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "Hello", " ", "line"],
                    [yuio.term.Color.NONE, "break"],
                ],
            ),
            (
                [yuio.term.Color.STYLE_BOLD, yuio.term.Color.NONE],
                15,
                {},
                [
                    [
                        yuio.term.Color.NONE,
                        yuio.term.Color.STYLE_BOLD,
                        yuio.term.Color.NONE,
                    ],
                ],
            ),
            (
                [yuio.term.Color.STYLE_BOLD, "", yuio.term.Color.NONE],
                15,
                {},
                [
                    [
                        yuio.term.Color.NONE,
                        yuio.term.Color.STYLE_BOLD,
                        yuio.term.Color.NONE,
                    ],
                ],
            ),
            (
                ["break\n", yuio.term.Color.NONE],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "break"],
                    [yuio.term.Color.NONE],
                ],
            ),
            (
                ["break\n"],
                15,
                {},
                [
                    [yuio.term.Color.NONE, "break"],
                    [yuio.term.Color.NONE],
                ],
            ),
            (
                [
                    "usage: app.py train [-h] [-v] [--force-color | --force-no-color] [-o {path}] <data>"
                ],
                100,
                {},
                [
                    [
                        yuio.term.Color.NONE,
                        "usage:",
                        " ",
                        "app.py",
                        " ",
                        "train",
                        " ",
                        "[-h]",
                        " ",
                        "[-v]",
                        " ",
                        "[--force-",
                        "color",
                        " ",
                        "|",
                        " ",
                        "--force-",
                        "no-",
                        "color]",
                        " ",
                        "[-o",
                        " ",
                        "{path}]",
                        " ",
                        "<data>",
                    ]
                ],
            ),
            (
                [
                    yuio.term.Color.FORE_MAGENTA,
                    "usage: ",
                    yuio.term.Color.NONE,
                    yuio.term.Color.NONE,
                    "app.py train",
                    yuio.term.Color.NONE,
                    yuio.term.Color.NONE,
                    " ",
                    yuio.term.Color.NONE,
                    "[",
                    yuio.term.Color.FORE_BLUE,
                    "-h",
                    yuio.term.Color.NONE,
                    "]",
                    yuio.term.Color.NONE,
                    " ",
                    yuio.term.Color.NONE,
                    "[",
                    yuio.term.Color.FORE_BLUE,
                    "-v",
                    yuio.term.Color.NONE,
                    "]",
                    yuio.term.Color.NONE,
                    " [",
                    yuio.term.Color.NONE,
                    yuio.term.Color.FORE_BLUE,
                    "--force-color",
                    yuio.term.Color.NONE,
                    " | ",
                    yuio.term.Color.NONE,
                    yuio.term.Color.FORE_BLUE,
                    "--force-no-color",
                    yuio.term.Color.NONE,
                    "] ",
                    yuio.term.Color.NONE,
                    "[",
                    yuio.term.Color.FORE_BLUE,
                    "-o",
                    yuio.term.Color.NONE,
                    " ",
                    yuio.term.Color.FORE_MAGENTA,
                    "",
                    yuio.term.Color.NONE,
                    "{",
                    yuio.term.Color.FORE_MAGENTA,
                    "path",
                    yuio.term.Color.NONE,
                    "}",
                    yuio.term.Color.FORE_MAGENTA,
                    "",
                    yuio.term.Color.NONE,
                    "]",
                    yuio.term.Color.NONE,
                    " ",
                    yuio.term.Color.NONE,
                    yuio.term.Color.FORE_MAGENTA,
                    "",
                    yuio.term.Color.NONE,
                    "<",
                    yuio.term.Color.FORE_MAGENTA,
                    "data",
                    yuio.term.Color.NONE,
                    ">",
                    yuio.term.Color.FORE_MAGENTA,
                    "",
                ],
                100,
                {},
                [
                    [
                        yuio.term.Color.NONE,
                        yuio.term.Color.FORE_MAGENTA,
                        "usage:",
                        " ",
                        yuio.term.Color.NONE,
                        "app.py",
                        " ",
                        "train",
                        " ",
                        "[",
                        yuio.term.Color.FORE_BLUE,
                        "-h",
                        yuio.term.Color.NONE,
                        "]",
                        " ",
                        "[",
                        yuio.term.Color.FORE_BLUE,
                        "-v",
                        yuio.term.Color.NONE,
                        "]",
                        " ",
                        "[",
                        yuio.term.Color.FORE_BLUE,
                        "--force-",
                        "color",
                        yuio.term.Color.NONE,
                        " ",
                        "|",
                        " ",
                        yuio.term.Color.FORE_BLUE,
                        "--force-",
                        "no-",
                        "color",
                        yuio.term.Color.NONE,
                        "]",
                        " ",
                        "[",
                        yuio.term.Color.FORE_BLUE,
                        "-o",
                        yuio.term.Color.NONE,
                        " ",
                        yuio.term.Color.FORE_MAGENTA,
                        yuio.term.Color.NONE,
                        "{",
                        yuio.term.Color.FORE_MAGENTA,
                        "path",
                        yuio.term.Color.NONE,
                        "}",
                        yuio.term.Color.FORE_MAGENTA,
                        yuio.term.Color.NONE,
                        "]",
                        " ",
                        yuio.term.Color.FORE_MAGENTA,
                        yuio.term.Color.NONE,
                        "<",
                        yuio.term.Color.FORE_MAGENTA,
                        "data",
                        yuio.term.Color.NONE,
                        ">",
                        yuio.term.Color.FORE_MAGENTA,
                    ]
                ],
            ),
            (
                ["single string"],
                100,
                {"first_line_indent": ">>"},
                [
                    [
                        yuio.term.Color.NONE,
                        ">>",
                        yuio.term.Color.NONE,
                        "single",
                        " ",
                        "string",
                    ],
                ],
            ),
            (
                ["single string"],
                100,
                {"continuation_indent": ">>"},
                [
                    [yuio.term.Color.NONE, "single", " ", "string"],
                ],
            ),
            (
                ["single string"],
                13,
                {},
                [
                    [yuio.term.Color.NONE, "single", " ", "string"],
                ],
            ),
            (
                ["single string"],
                13,
                {"first_line_indent": ">>", "continuation_indent": ".."},
                [
                    [yuio.term.Color.NONE, ">>", yuio.term.Color.NONE, "single"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "string"],
                ],
            ),
            (
                ["foo bar baz"],
                8,
                {"first_line_indent": ">>>", "continuation_indent": "|"},
                [
                    [yuio.term.Color.NONE, ">>>", yuio.term.Color.NONE, "foo"],
                    [
                        yuio.term.Color.NONE,
                        "|",
                        yuio.term.Color.NONE,
                        "bar",
                        " ",
                        "baz",
                    ],
                ],
            ),
            (
                ["word werywerylongunbreakableword"],
                8,
                {"first_line_indent": ">>", "continuation_indent": ".."},
                [
                    [yuio.term.Color.NONE, ">>", yuio.term.Color.NONE, "word"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "werywe"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "rylong"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "unbrea"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "kablew"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "ord"],
                ],
            ),
            (
                ["werywerylongunbreakableword"],
                8,
                {"first_line_indent": ">>>", "continuation_indent": "."},
                [
                    [yuio.term.Color.NONE, ">>>", yuio.term.Color.NONE, "weryw"],
                    [yuio.term.Color.NONE, ".", yuio.term.Color.NONE, "erylong"],
                    [yuio.term.Color.NONE, ".", yuio.term.Color.NONE, "unbreak"],
                    [yuio.term.Color.NONE, ".", yuio.term.Color.NONE, "ablewor"],
                    [yuio.term.Color.NONE, ".", yuio.term.Color.NONE, "d"],
                ],
            ),
            (
                ["single string", "\nnext string"],
                13,
                {"first_line_indent": ">>", "continuation_indent": ".."},
                [
                    [yuio.term.Color.NONE, ">>", yuio.term.Color.NONE, "single"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "string"],
                    [
                        yuio.term.Color.NONE,
                        "..",
                        yuio.term.Color.NONE,
                        "next",
                        " ",
                        "string",
                    ],
                ],
            ),
            (
                ["a\n\nb"],
                13,
                {"first_line_indent": ">>", "continuation_indent": ".."},
                [
                    [yuio.term.Color.NONE, ">>", yuio.term.Color.NONE, "a"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "b"],
                ],
            ),
            (
                ["a\n"],
                13,
                {"first_line_indent": ">>", "continuation_indent": ".."},
                [
                    [yuio.term.Color.NONE, ">>", yuio.term.Color.NONE, "a"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE],
                ],
            ),
            (
                ["\na"],
                13,
                {"first_line_indent": ">>", "continuation_indent": ".."},
                [
                    [yuio.term.Color.NONE, ">>", yuio.term.Color.NONE],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "a"],
                ],
            ),
            (
                ["\nwerywerylongunbreakableword"],
                13,
                {"first_line_indent": ">>", "continuation_indent": ".."},
                [
                    [yuio.term.Color.NONE, ">>", yuio.term.Color.NONE],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "werywerylon"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "gunbreakabl"],
                    [yuio.term.Color.NONE, "..", yuio.term.Color.NONE, "eword"],
                ],
            ),
            (
                [yuio.term.Color.FORE_BLUE, "single string"],
                13,
                {
                    "first_line_indent": yuio.term.ColorizedString(
                        [yuio.term.Color.FORE_MAGENTA, ">>"]
                    ),
                    "continuation_indent": yuio.term.ColorizedString(
                        [yuio.term.Color.FORE_BLUE, ".."]
                    ),
                },
                [
                    [
                        yuio.term.Color.NONE,
                        yuio.term.Color.FORE_MAGENTA,
                        ">>",
                        yuio.term.Color.NONE,
                        yuio.term.Color.FORE_BLUE,
                        "single",
                    ],
                    [
                        yuio.term.Color.NONE,
                        yuio.term.Color.FORE_BLUE,
                        "..",
                        yuio.term.Color.FORE_BLUE,
                        "string",
                    ],
                ],
            ),
        ],
    )
    def test_wrap(self, text, width, kwargs, expect):
        wrapped = yuio.term.ColorizedString(text).wrap(width, **kwargs)
        raw = [line._parts for line in wrapped]
        assert raw == expect

    @pytest.mark.parametrize(
        "text,first_line_indent,continuation_indent,expect",
        [
            (
                yuio.term.ColorizedString(),
                yuio.term.ColorizedString(),
                yuio.term.ColorizedString(),
                [],
            ),
            (
                yuio.term.ColorizedString(""),
                yuio.term.ColorizedString(),
                yuio.term.ColorizedString(),
                [],
            ),
            (
                yuio.term.ColorizedString("abc"),
                yuio.term.ColorizedString(),
                yuio.term.ColorizedString(),
                [yuio.term.Color.NONE, "abc"],
            ),
            (
                yuio.term.ColorizedString("abc\n"),
                yuio.term.ColorizedString(),
                yuio.term.ColorizedString(),
                [yuio.term.Color.NONE, "abc\n"],
            ),
            (
                yuio.term.ColorizedString(),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                [],
            ),
            (
                yuio.term.ColorizedString(""),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                [],
            ),
            (
                yuio.term.ColorizedString("abc"),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                ["1", yuio.term.Color.NONE, "abc"],
            ),
            (
                yuio.term.ColorizedString("abc\n"),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                ["1", yuio.term.Color.NONE, "abc\n"],
            ),
            (
                yuio.term.ColorizedString(["abc", "def"]),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                ["1", yuio.term.Color.NONE, "abc", "def"],
            ),
            (
                yuio.term.ColorizedString("abc\ndef"),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                ["1", yuio.term.Color.NONE, "abc\n", "2", yuio.term.Color.NONE, "def"],
            ),
            (
                yuio.term.ColorizedString(["abc\n", "def"]),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                ["1", yuio.term.Color.NONE, "abc\n", "2", yuio.term.Color.NONE, "def"],
            ),
            (
                yuio.term.ColorizedString(["abc", "\n", "def"]),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                [
                    "1",
                    yuio.term.Color.NONE,
                    "abc",
                    "\n",
                    "2",
                    yuio.term.Color.NONE,
                    "def",
                ],
            ),
            (
                yuio.term.ColorizedString("abc\ndef\n"),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString("2"),
                [
                    "1",
                    yuio.term.Color.NONE,
                    "abc\n",
                    "2",
                    yuio.term.Color.NONE,
                    "def\n",
                ],
            ),
            (
                yuio.term.ColorizedString("abc\ndef"),
                yuio.term.ColorizedString("1"),
                yuio.term.ColorizedString(),
                ["1", yuio.term.Color.NONE, "abc\n", yuio.term.Color.NONE, "def"],
            ),
            (
                yuio.term.ColorizedString("abc\ndef"),
                yuio.term.ColorizedString(),
                yuio.term.ColorizedString("2"),
                [yuio.term.Color.NONE, "abc\n", "2", yuio.term.Color.NONE, "def"],
            ),
            (
                yuio.term.ColorizedString(["abc\ndef"]),
                yuio.term.ColorizedString([yuio.term.Color.FORE_RED, "1"]),
                yuio.term.ColorizedString([yuio.term.Color.FORE_BLUE, "2"]),
                [
                    yuio.term.Color.FORE_RED,
                    "1",
                    yuio.term.Color.NONE,
                    "abc\n",
                    yuio.term.Color.FORE_BLUE,
                    "2",
                    yuio.term.Color.NONE,
                    "def",
                ],
            ),
            (
                yuio.term.ColorizedString([yuio.term.Color.FORE_YELLOW, "abc\ndef"]),
                yuio.term.ColorizedString([yuio.term.Color.FORE_RED, "1"]),
                yuio.term.ColorizedString([yuio.term.Color.FORE_BLUE, "2"]),
                [
                    yuio.term.Color.FORE_RED,
                    "1",
                    yuio.term.Color.FORE_YELLOW,
                    "abc\n",
                    yuio.term.Color.NONE,
                    yuio.term.Color.FORE_BLUE,
                    "2",
                    yuio.term.Color.FORE_YELLOW,
                    "def",
                ],
            ),
            (
                yuio.term.ColorizedString(
                    ["abc\n", yuio.term.Color.FORE_YELLOW, "def\nhig"]
                ),
                yuio.term.ColorizedString([yuio.term.Color.FORE_RED, "1"]),
                yuio.term.ColorizedString([yuio.term.Color.FORE_BLUE, "2"]),
                [
                    yuio.term.Color.FORE_RED,
                    "1",
                    yuio.term.Color.NONE,
                    "abc\n",
                    yuio.term.Color.FORE_BLUE,
                    "2",
                    yuio.term.Color.FORE_YELLOW,
                    "def\n",
                    yuio.term.Color.NONE,
                    yuio.term.Color.FORE_BLUE,
                    "2",
                    yuio.term.Color.FORE_YELLOW,
                    "hig",
                ],
            ),
            (
                yuio.term.ColorizedString(
                    ["abc", yuio.term.Color.FORE_YELLOW, "\ndef\nhig"]
                ),
                yuio.term.ColorizedString([yuio.term.Color.FORE_RED, "1"]),
                yuio.term.ColorizedString([yuio.term.Color.FORE_BLUE, "2"]),
                [
                    yuio.term.Color.FORE_RED,
                    "1",
                    yuio.term.Color.NONE,
                    "abc",
                    yuio.term.Color.FORE_YELLOW,
                    "\n",
                    yuio.term.Color.NONE,
                    yuio.term.Color.FORE_BLUE,
                    "2",
                    yuio.term.Color.FORE_YELLOW,
                    "def\n",
                    yuio.term.Color.NONE,
                    yuio.term.Color.FORE_BLUE,
                    "2",
                    yuio.term.Color.FORE_YELLOW,
                    "hig",
                ],
            ),
        ],
    )
    def test_indent(self, text, first_line_indent, continuation_indent, expect):
        indented = text.indent(first_line_indent, continuation_indent)
        assert indented._parts == expect


class MockOStream(io.StringIO):
    def __init__(
        self,
        out: "MockIStream",
        tty: bool,
        should_query_osc: bool,
        osc_response: str | None,
    ):
        super().__init__()
        self.__out = out
        self.__tty = tty
        self.__should_query_osc = should_query_osc
        self.__osc_response = osc_response

    def isatty(self) -> bool:
        return self.__tty

    _OSC_Q = (
        "\x1b]10;?\x1b\\"
        "\x1b]11;?"
        "\x1b\\"
        "\x1b]4;0;?\x1b\\"
        "\x1b]4;1;?\x1b\\"
        "\x1b]4;2;?\x1b\\"
        "\x1b]4;3;?\x1b\\"
        "\x1b]4;4;?\x1b\\"
        "\x1b]4;5;?\x1b\\"
        "\x1b]4;6;?\x1b\\"
        "\x1b]4;7;?\x1b\\"
        "\x1b[c"
    )

    _OSC_R = (
        "\x1b]10;rgb:bfbf/bfbf/bfbf\x1b\\"
        "\x1b]11;rgb:0000/0000/0000\x1b\\"
        "\x1b]4;0;rgb:0000/0000/0000\x1b\\"
        "\x1b]4;1;rgb:d4d4/2c2c/3a3a\x1b\\"
        "\x1b]4;2;rgb:1c1c/a8a8/0000\x1b\\"
        "\x1b]4;3;rgb:c0c0/a0a0/0000\x1b\\"
        "\x1b]4;4;rgb:0000/5d5d/ffff\x1b\\"
        "\x1b]4;5;rgb:b1b1/4848/c6c6\x1b\\"
        "\x1b]4;6;rgb:0000/a8a8/9a9a\x1b\\"
        "\x1b]4;7;rgb:bfbf/bfbf/bfbf\x1b\\"
        "\x1b[?c"
    )

    def write(self, s) -> int:
        if s == self._OSC_Q and self.__osc_response != "":
            if not self.__should_query_osc:
                raise RuntimeError("terminal is not supposed to query OSC")
            self._add_to_out(self.__osc_response or self._OSC_R)
        ret = super().write(s)
        return ret

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def _add_to_out(self, s):
        pos = self.__out.tell()
        self.__out.seek(0, io.SEEK_END)
        self.__out.write(s)
        self.__out.seek(pos)
        self.__out.flush()


class MockIStream(io.StringIO):
    def __init__(
        self,
        tty: bool,
    ):
        super().__init__()
        self.__tty = tty

    def isatty(self) -> bool:
        return self.__tty

    def read(self, size=None):
        raise RuntimeError("term not supposed to use stream.read")

    def readline(self, size=None):
        raise RuntimeError("term not supposed to use stream.readline")

    def readlines(self, hint=-1):
        raise RuntimeError("term not supposed to use stream.readlines")

    def read_keycode(self):
        return super().read()


@contextlib.contextmanager
def mock_term_io(
    i_tty: bool = False,
    o_tty: bool = False,
    should_query_osc: bool = False,
    is_foreground: bool = False,
    env: dict[str, str] = {},
    args: list[str] = [],
    osc_response: str | None = "",
    enable_vt_processing: bool = False,
):
    istream = MockIStream(i_tty)
    ostream = MockOStream(istream, o_tty, should_query_osc, osc_response)

    old_env = os.environ.copy()
    os.environ.clear()
    os.environ.update(env)

    old_args = sys.argv[1:]
    sys.argv[1:] = args

    old_read_keycode, yuio.term._read_keycode = (
        yuio.term._read_keycode,
        istream.read_keycode,
    )
    old_is_foreground, yuio.term._is_foreground = (
        yuio.term._is_foreground,
        lambda _: is_foreground,
    )
    old_is_interactive_input, yuio.term._is_interactive_input = (
        yuio.term._is_interactive_input,
        lambda _: i_tty,
    )
    old_enter_raw_mode, yuio.term._enter_raw_mode = (
        yuio.term._enter_raw_mode,
        lambda *_, **__: contextlib.nullcontext(),
    )
    old_enable_vt_processing, yuio.term._enable_vt_processing = (
        yuio.term._enable_vt_processing,
        lambda ostream: enable_vt_processing,
    )
    old_flush_input_buffer, yuio.term._flush_input_buffer = (
        yuio.term._flush_input_buffer,
        lambda: None,
    )

    try:
        yield ostream, istream
    finally:
        os.environ.clear()
        os.environ.update(old_env)

        sys.argv[1:] = old_args

        yuio.term._read_keycode = old_read_keycode
        yuio.term._is_foreground = old_is_foreground
        yuio.term._is_interactive_input = old_is_interactive_input
        yuio.term._enter_raw_mode = old_enter_raw_mode
        yuio.term._enable_vt_processing = old_enable_vt_processing
        yuio.term._flush_input_buffer = old_flush_input_buffer


term_colors = yuio.term.TerminalColors(
    background=yuio.term.ColorValue.from_hex("#000000"),
    foreground=yuio.term.ColorValue.from_hex("#BFBFBF"),
    black=yuio.term.ColorValue.from_hex("#000000"),
    red=yuio.term.ColorValue.from_hex("#D42C3A"),
    green=yuio.term.ColorValue.from_hex("#1CA800"),
    yellow=yuio.term.ColorValue.from_hex("#C0A000"),
    blue=yuio.term.ColorValue.from_hex("#005DFF"),
    magenta=yuio.term.ColorValue.from_hex("#B148C6"),
    cyan=yuio.term.ColorValue.from_hex("#00A89A"),
    white=yuio.term.ColorValue.from_hex("#BFBFBF"),
    lightness=yuio.term.Lightness.DARK,
)


class TestTerm:
    @pytest.mark.parametrize(
        "level,ansi,ansi_256,ansi_true",
        [
            (yuio.term.ColorSupport.ANSI, True, False, False),
            (yuio.term.ColorSupport.ANSI_256, True, True, False),
            (yuio.term.ColorSupport.ANSI_TRUE, True, True, True),
        ],
    )
    def test_color_support(self, level, ansi, ansi_256, ansi_true):
        term = yuio.term.Term(None, None, color_support=level)  # type: ignore
        assert term.has_colors == ansi
        assert term.has_colors_256 == ansi_256
        assert term.has_colors_true == ansi_true

    @pytest.mark.parametrize(
        "level,move,query",
        [
            (yuio.term.InteractiveSupport.NONE, False, False),
            (yuio.term.InteractiveSupport.MOVE_CURSOR, True, False),
            (yuio.term.InteractiveSupport.FULL, True, True),
        ],
    )
    def test_interactive_support(self, level, move, query):
        term = yuio.term.Term(None, None, color_support=yuio.term.ColorSupport.ANSI, interactive_support=level)  # type: ignore
        assert term.can_move_cursor == move
        assert term.can_query_terminal == term.is_fully_interactive == query
        term = yuio.term.Term(None, None, color_support=yuio.term.ColorSupport.NONE, interactive_support=level)  # type: ignore
        assert (
            term.can_move_cursor
            == term.can_query_terminal
            == term.is_fully_interactive
            == False
        )

    @pytest.mark.skipif(os.name == "nt", reason="this test is for unix only")
    @pytest.mark.parametrize(
        "kwargs,expected_term",
        [
            (
                {},
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {"env": {"TERM": "xterm"}},
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm"},
                    "i_tty": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm"},
                    "o_tty": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm"},
                    "i_tty": True,
                    "o_tty": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm"},
                    "is_foreground": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm"},
                    "i_tty": True,
                    "is_foreground": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm"},
                    "o_tty": True,
                    "is_foreground": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.MOVE_CURSOR,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm", "COLORTERM": "yes"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "should_query_osc": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_256,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm", "COLORTERM": "truecolor"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "should_query_osc": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,  # OSC query got no response
                },
            ),
            (
                {
                    "env": {"TERM": "xterm", "COLORTERM": "yes"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "should_query_osc": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_256,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,  # kbhit responds, but no OSC result
                },
            ),
            (
                {
                    "env": {"TERM": "xterm", "COLORTERM": "yes"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "should_query_osc": True,
                    "osc_response": "\x1b[?c",
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_256,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,  # kbhit responds, but OSC is not properly supported
                },
            ),
            (
                {
                    "env": {"TERM": "xterm", "COLORTERM": "yes"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "should_query_osc": True,
                    "osc_response": None,  # default response
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_256,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": term_colors,  # Got the response!
                },
            ),
            (
                {
                    "env": {"TERM": "xterm", "COLORTERM": "yes"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "args": ["--no-color"],
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"TERM": "xterm", "COLORTERM": "yes", "FORCE_NO_COLOR": "1"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "args": ["--force-color"],
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {"env": {"FORCE_COLOR": "1"}},
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
        ],
    )
    def test_capabilities_estimation(self, kwargs, expected_term):
        with mock_term_io(**kwargs) as streams:
            ostream, istream = streams
            term = yuio.term.get_term_from_stream(ostream, istream)
            expected = yuio.term.Term(ostream, istream, **expected_term)
            assert term == expected

    @pytest.mark.skipif(os.name != "nt", reason="this test is for windows only")
    @pytest.mark.parametrize(
        "kwargs,expected_term",
        [
            (
                {},
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {"enable_vt_processing": True},
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "i_tty": True,
                    "enable_vt_processing": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "o_tty": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "o_tty": True,
                    "enable_vt_processing": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "i_tty": True,
                    "o_tty": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "i_tty": True,
                    "o_tty": True,
                    "enable_vt_processing": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "is_foreground": True,
                    "enable_vt_processing": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "i_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "o_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.MOVE_CURSOR,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                    "should_query_osc": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,  # OSC query got no response
                },
            ),
            (
                {
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                    "should_query_osc": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,  # kbhit responds, but no OSC result
                },
            ),
            (
                {
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                    "should_query_osc": True,
                    "osc_response": "\x1b[?c",
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": None,  # kbhit responds, but OSC is not properly supported
                },
            ),
            (
                {
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                    "should_query_osc": True,
                    "osc_response": None,  # default response
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI_TRUE,
                    "interactive_support": yuio.term.InteractiveSupport.FULL,
                    "terminal_colors": term_colors,  # Got the response!
                },
            ),
            (
                {
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                    "args": ["--no-color"],
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "env": {"FORCE_NO_COLOR": "1"},
                    "i_tty": True,
                    "o_tty": True,
                    "is_foreground": True,
                    "enable_vt_processing": True,
                },
                {
                    "color_support": yuio.term.ColorSupport.NONE,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {
                    "args": ["--force-color"],
                },
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
            (
                {"env": {"FORCE_COLOR": "1"}},
                {
                    "color_support": yuio.term.ColorSupport.ANSI,
                    "interactive_support": yuio.term.InteractiveSupport.NONE,
                    "terminal_colors": None,
                },
            ),
        ],
    )
    def test_capabilities_estimation_windows(self, kwargs, expected_term):
        with mock_term_io(**kwargs) as streams:
            ostream, istream = streams
            term = yuio.term.get_term_from_stream(ostream, istream)
            expected = yuio.term.Term(ostream, istream, **expected_term)
            assert term == expected



import io
import sys

import pytest

import yuio.io
import yuio.parse
import yuio.term
import yuio.theme
import yuio.widget
from yuio import _t

from .conftest import IOMocker, RcCompare


@pytest.fixture()
def enable_bg_updates() -> bool:
    return False


@pytest.fixture(autouse=True)
def setup(
    term: yuio.term.Term,
    theme: yuio.theme.Theme,
    monkeypatch: pytest.MonkeyPatch,
    enable_bg_updates: bool,
    width: int,
):
    theme.spinner_pattern = "⣿"
    theme.set_color("code", "magenta")
    theme.set_color("msg/decoration", "magenta")
    theme.set_color("msg/text:heading", "bold")
    theme.set_color("msg/text:question", "blue")
    theme.set_color("msg/text:error", "red")
    theme.set_color("msg/text:warning", "yellow")
    theme.set_color("msg/text:success", "green")
    theme.set_color("msg/text:info", "cyan")
    theme.progress_bar_width = 5

    io_manager = yuio.io._IoManager(term, theme, enable_bg_updates=enable_bg_updates)
    monkeypatch.setattr("yuio.io._IO_MANAGER", io_manager)
    io_manager.formatter.width = width

    yield

    io_manager.stop()


class TestSetup:
    @pytest.fixture(autouse=True)
    def setup(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io._IO_MANAGER", None)

    @pytest.mark.skip(reason="TODO")
    def test_simple(self):
        pass

    @pytest.mark.skip(reason="TODO")
    def test_term(self):
        pass

    @pytest.mark.skip(reason="TODO")
    def test_theme(self):
        pass

    @pytest.mark.skip(reason="TODO")
    def test_theme_callable(self):
        pass

    @pytest.mark.skip(reason="TODO")
    def test_term_theme(self):
        pass

    @pytest.mark.skip(reason="TODO")
    def test_term_theme_callable(self):
        pass

    @pytest.mark.skip(reason="TODO")
    def test_wrap_streams(self):
        pass

    @pytest.mark.skip(reason="TODO")
    def test_wrap_streams_non_interactive(self):
        pass


class TestMessage:
    def test_info(self, ostream: io.StringIO):
        yuio.io.info("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "cccccccc            ",
            ],
        )

    def test_info_wrap(self, ostream: io.StringIO):
        yuio.io.info("foo bar baz foo bar baz foo bar baz!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar baz foo bar ",
                "baz foo bar baz!    ",
            ],
            [
                "ccccccccccccccccccc ",
                "cccccccccccccccc    ",
            ],
        )

    def test_info_args(self, ostream: io.StringIO):
        yuio.io.info("%s %s!", "foo", "bar")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "cccccccc            ",
            ],
        )

    def test_info_color_tags(self, ostream: io.StringIO):
        yuio.io.info("`%s` <c bold>%s</c>!", "<c green>foo</c>", "bar")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "<c green>foo</c> bar",
                "!                   ",
            ],
            [
                "mmmmmmmmmmmmmmmmcCCC",
                "c                   ",
            ],
        )

    def test_warning(self, ostream: io.StringIO):
        yuio.io.warning("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "yyyyyyyy            ",
            ],
        )

    def test_error(self, ostream: io.StringIO):
        yuio.io.error("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "rrrrrrrr            ",
            ],
        )

    def test_error_with_tb(self, ostream: io.StringIO, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr(
            "traceback.format_exception", lambda ty, val, tb: [f"{ty.__name__}: {val}"]
        )

        try:
            raise RuntimeError("something happened")
        except RuntimeError:
            yuio.io.error_with_tb("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
                "  RuntimeError: some",
                "thing happened      ",
            ],
            [
                "rrrrrrrr            ",
                "                    ",
                "                    ",
            ],
        )

    def test_error_with_tb_manual(
        self, ostream: io.StringIO, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setattr(
            "traceback.format_exception", lambda ty, val, tb: [f"{ty.__name__}: {val}"]
        )

        try:
            raise RuntimeError("something happened")
        except RuntimeError as e:
            yuio.io.error_with_tb("foo bar!", exc_info=(TypeError, e, e.__traceback__))
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
                "  TypeError: somethi",
                "ng happened         ",
            ],
            [
                "rrrrrrrr            ",
                "                    ",
                "                    ",
            ],
        )

    def test_success(self, ostream: io.StringIO):
        yuio.io.success("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "gggggggg            ",
            ],
        )

    def test_heading(self, ostream: io.StringIO):
        yuio.io.heading("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "⣿ foo bar!          ",
            ],
            [
                "mm########          ",
            ],
        )

    def test_md(self, ostream: io.StringIO):
        yuio.io.md("# Foo!\n\n- bar\n- baz")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "⣿ Foo!              ",
                "                    ",
                "•   bar             ",
                "                    ",
                "•   baz             ",
            ],
            [
                "mm####              ",
                "                    ",
                "mmmm                ",
                "                    ",
                "mmmm                ",
            ],
        )

    def test_br(self, ostream: io.StringIO):
        yuio.io.info("foo")
        yuio.io.br()
        yuio.io.info("bar")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo                 ",
                "                    ",
                "bar                 ",
            ],
            [
                "ccc                 ",
                "                    ",
                "ccc                 ",
            ],
        )

    def test_raw(self, ostream: io.StringIO):
        yuio.io.raw(
            yuio.term.ColorizedString(["foo, ", yuio.term.Color.FORE_RED, "bar"])
        )

        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo, bar            ",
            ],
            [
                "     rrr            ",
            ],
        )


class TestAsk:
    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Hello?              ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("Hii~")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "Hii~"

    def test_empty(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Hello?              ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)
        io_mocker.expect_screen(
            [
                "Hello?              ",
                ">                   ",
                "▲ Error: input is   ",
                "required.           ",
                "f1 help             ",
            ],
        )
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "123"

    def test_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "What's your deal?   ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("meow =^..^=")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("What's your %s?", "deal") == "meow =^..^="

    def test_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter something:    ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Enter something") == "123"

    def test_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter something:    ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Enter something:") == "123"

    def test_default(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q?                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "{default}"

    def test_default_overridden(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q?                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("foo!")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "foo!"

    def test_default_optional(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q?                  ",
                "> <none>            ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default=None) is None

    def test_default_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q", default="{default}") == "{default}"

    def test_default_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q:", default="{default}") == "{default}"

    def test_input_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                ">                   ",
                "<text> • f1 help    ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.F1)
        io_mocker.expect_screen(
            [
                "Input Format        ",
                "                    ",
                "            <text>  ",
                "                    ",
                ":                   ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ESCAPE)
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q:", input_description="<text>") == "123"

    def test_default_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                "> <default>         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert (
                yuio.io.ask("Q:", default="{default}", default_description="<default>")
                == "{default}"
            )

    def test_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Are you there?      ",
                "> no                ",
                "  yes               ",
                "f1 help             ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.TAB)
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Are you there?", parser=yuio.parse.Bool())

    def test_type_hint(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Are you there?      ",
                "> no                ",
                "  yes               ",
                "f1 help             ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.TAB)
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask[bool]("Are you there?")

    def test_type_hint_and_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter some numbers: ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123 456")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask[bool](
                "Enter some numbers", parser=yuio.parse.List(yuio.parse.Int())  # type: ignore
            ) == [123, 456]


class TestAskNonInteractive:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            interactive_support=yuio.term.InteractiveSupport.MOVE_CURSOR,
        )

    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Hello?              ",
            ],
        )
        io_mocker.expect_istream_readline("Hii~")

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "Hii~"

    def test_empty(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Hello?              ",
            ],
        )
        io_mocker.expect_istream_readline("\n")
        io_mocker.expect_screen(
            [
                "Hello?              ",
                "Input is required.  ",
                "Hello?              ",
            ],
        )
        io_mocker.expect_istream_readline("Hii~\n")

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "Hii~"

    def test_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "What's your deal?   ",
            ],
        )
        io_mocker.expect_istream_readline("meow =^..^=\n")

        with io_mocker.mock():
            assert yuio.io.ask("What's your %s?", "deal") == "meow =^..^="

    def test_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter something:    ",
            ],
        )
        io_mocker.expect_istream_readline("123\n")

        with io_mocker.mock():
            assert yuio.io.ask("Enter something") == "123"

    def test_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter something:    ",
            ],
        )
        io_mocker.expect_istream_readline("123\n")

        with io_mocker.mock():
            assert yuio.io.ask("Enter something:") == "123"

    def test_default(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q? [{default}]      ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "{default}"

    def test_default_overridden(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q? [{default}]      ",
            ],
        )
        io_mocker.expect_istream_readline("foo!\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "foo!"

    def test_default_optional(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q? [<none>]         ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default=None) is None

    def test_default_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q [{default}]:      ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q", default="{default}") == "{default}"

    def test_default_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q [{default}]:      ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q:", default="{default}") == "{default}"

    def test_input_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q (<text>):         ",
            ],
        )
        io_mocker.expect_istream_readline("123\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q:", input_description="<text>") == "123"

    def test_default_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q [<default>]:      ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert (
                yuio.io.ask("Q:", default="{default}", default_description="<default>")
                == "{default}"
            )

    def test_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Are you there? (yes|",
                "no)                 ",
            ],
        )
        io_mocker.expect_istream_readline("what?\n")
        io_mocker.expect_screen(
            [
                "Are you there? (yes|",
                "no) what?           ",
                "Error: can't parse  ",
                "'what?', enter      ",
                "either 'yes' or     ",
                "'no'.               ",
                "Are you there? (yes|",
                "no)                 ",
            ],
        )
        io_mocker.expect_istream_readline("y\n")

        with io_mocker.mock():
            assert yuio.io.ask("Are you there?", parser=yuio.parse.Bool())

    def test_parser_hint(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Are you there? (yes|",
                "no)                 ",
            ],
        )
        io_mocker.expect_istream_readline("what?\n")
        io_mocker.expect_screen(
            [
                "Are you there? (yes|",
                "no) what?           ",
                "Error: can't parse  ",
                "'what?', enter      ",
                "either 'yes' or     ",
                "'no'.               ",
                "Are you there? (yes|",
                "no)                 ",
            ],
        )
        io_mocker.expect_istream_readline("y\n")

        with io_mocker.mock():
            assert yuio.io.ask[bool]("Are you there?")

    def test_type_hint_and_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter some numbers (",
                "<int>[ <int>[ ...]])",
                ":                   ",
            ],
        )
        io_mocker.expect_istream_readline("123 456\n")

        with io_mocker.mock():
            assert yuio.io.ask[bool](
                "Enter some numbers", parser=yuio.parse.List(yuio.parse.Int())  # type: ignore
            ) == [123, 456]


class TestAskUnreadable:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        istream.readable = lambda: False
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            interactive_support=yuio.term.InteractiveSupport.NONE,
        )

    def test_default(self, io_mocker: IOMocker):
        with io_mocker.mock():
            assert yuio.io.ask("Meow?", default="Meow!") == "Meow!"

    def test_no_default(self, io_mocker: IOMocker):
        with io_mocker.mock():
            with pytest.raises(
                yuio.io.UserIoError, match="non-interactive environment"
            ):
                yuio.io.ask("Meow?")


class TestWaitForUser:
    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Press enter to      ",
                "continue            ",
            ]
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            yuio.io.wait_for_user()

    def test_msg(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Slam that Enter     ",
                "button!             ",
            ]
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that Enter button!")

    def test_msg_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Slam that Enter     ",
                "button!             ",
            ]
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that %s button!", "Enter")


class TestWaitForUserNonInteractive:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            interactive_support=yuio.term.InteractiveSupport.MOVE_CURSOR,
        )

    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Press enter to conti",
                "nue                 ",
            ]
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            yuio.io.wait_for_user()

    def test_add_space(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Press enter to conti",
                "nue                 ",
            ]
        )
        io_mocker.expect_istream_readline(".\n")
        io_mocker.expect_screen(
            [
                "Press enter to conti",
                "nue .               ",
            ]
        )

        with io_mocker.mock():
            yuio.io.wait_for_user()

    def test_dont_add_space(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Ends with space     ",
            ]
        )
        io_mocker.expect_istream_readline(".\n")
        io_mocker.expect_screen(
            [
                "Ends with space .   ",
            ]
        )

        with io_mocker.mock():
            yuio.io.wait_for_user("Ends with space ")

    def test_msg(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Slam that Enter butt",
                "on!                 ",
            ]
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that Enter button!")

    def test_msg_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Slam that Enter butt",
                "on!                 ",
            ]
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that %s button!", "Enter")


class TestWaitForUserUnreadable:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        istream.readable = lambda: False
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            interactive_support=yuio.term.InteractiveSupport.NONE,
        )

    def test_wait(self, io_mocker: IOMocker):
        io_mocker.expect_screen([])
        with io_mocker.mock():
            yuio.io.wait_for_user()


class TestDetectEditor:
    def test_env(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("EDITOR", "foobar")
        assert yuio.io.detect_editor() == "foobar"

    @pytest.mark.parametrize("editor", ["vi", "nano", "notepad.exe"])
    def test_which(self, editor: str, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("shutil.which", lambda exc: exc if exc == editor else None)
        assert yuio.io.detect_editor() == editor

    def test_fail(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.delenv("EDITOR", raising=False)
        monkeypatch.setattr("shutil.which", lambda exc: None)
        assert yuio.io.detect_editor() is None


class TestEdit:
    # no space after 'edited' because on windows,
    # cmd.exe adds space to the output (wtf?)
    _EDITOR = "echo edited>>"

    def test_simple(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: self._EDITOR)
        assert yuio.io.edit("foobar") == "foobaredited\n"

    def test_editor(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: None)
        assert yuio.io.edit("foobar", editor=self._EDITOR) == "foobaredited\n"

    def test_editor_error(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: "exit 1; cat")
        with pytest.raises(yuio.io.UserIoError, match="editing failed"):
            assert yuio.io.edit("foobar")

    def test_no_editor(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: None)
        with pytest.raises(yuio.io.UserIoError, match="can't detect an editor"):
            assert yuio.io.edit("foobar")

    def test_comments(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: self._EDITOR)
        assert yuio.io.edit("# foo\n  # bar\nbaz #") == "baz #edited\n"
        assert yuio.io.edit("foo\n#") == "foo\n"

    def test_comments_custom_marker(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: self._EDITOR)
        assert (
            yuio.io.edit("// foo\n  # bar\nbaz //", comment_marker="//")
            == "  # bar\nbaz //edited\n"
        )
        assert yuio.io.edit("foo\n//", comment_marker="//") == "foo\n"

    def test_comments_custom_marker_special_symbols(
        self, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: self._EDITOR)
        assert yuio.io.edit("a\nb\n[ab]", comment_marker="[ab]") == "a\nb\n"

    def test_file_removed(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda: "rm")
        assert yuio.io.edit("foo") == ""


class TestTask:
    @pytest.fixture()
    def width(self):
        return 40

    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task"):
                io_mocker.mark()

    def test_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task 1                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task 1 - done                         ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task `%s`", 1):
                io_mocker.mark()

    def test_print_while_in_task(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task 1                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "hello world!                            ",
                "⣿ task 1                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "hello world!                            ",
                "> task 1 - done                         ",
                "hello world 2!                          ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task `%s`", 1):
                io_mocker.mark()
                yuio.io.info("hello world!")
                io_mocker.mark()
            yuio.io.info("hello world 2!")

    def test_error_in_task(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - error                          ",
            ]
        )

        with io_mocker.mock():
            with pytest.raises(RuntimeError, match="eh..."):
                with yuio.io.Task("task"):
                    io_mocker.mark()
                    raise RuntimeError("eh...")

    def test_manual_error(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - error                          ",
            ]
        )

        with io_mocker.mock():
            task = yuio.io.Task("task")
            io_mocker.mark()
            task.error()

    def test_manual_done(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            task = yuio.io.Task("task")
            io_mocker.mark()
            task.done()

    def test_comment(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task - yaaay!                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.comment("yaaay!")
                io_mocker.mark()

    def test_comment_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task - yaaay 1!                       ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.comment("yaaay `%s`!", 1)
                io_mocker.mark()

    def test_comment_format_chached(self, io_mocker: IOMocker):
        class Fmt:
            i = 0

            def __str__(self):
                self.i += 1
                return str(self.i)

        fmt = Fmt()

        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task - yaaay 1!                       ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ task - 50.00% - yaaay 1!          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ task - 50.00% - yaaay 2!          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.comment("yaaay `%s`!", fmt)
                io_mocker.mark()
                task.progress(0.5)
                io_mocker.mark()
                task.comment("yaaay `%s`!", fmt)
                io_mocker.mark()

    @pytest.mark.parametrize(
        "args,kwargs,comment,expected",
        [
            (
                (0.3,),
                {},
                None,
                [
                    "■■□□□ task - 30.00%                     ",
                ],
            ),
            (
                (0.3,),
                {"unit": "whatever"},
                None,
                [
                    "■■□□□ task - 30.00%                     ",
                ],
            ),
            (
                (0.3,),
                {"ndigits": 0},
                None,
                [
                    "■■□□□ task - 30%                        ",
                ],
            ),
            (
                (11, 15),
                {},
                None,
                [
                    "■■■■□ task - 11/15                      ",
                ],
            ),
            (
                (10.5, 15),
                {},
                None,
                [
                    "■■■■□ task - 10.50/15.00                ",
                ],
            ),
            (
                (11, 15.0),
                {},
                None,
                [
                    "■■■■□ task - 11.00/15.00                ",
                ],
            ),
            (
                (10.5, 15),
                {"unit": "Kg"},
                None,
                [
                    "■■■■□ task - 10.50/15.00Kg              ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3},
                None,
                [
                    "■■■■□ task - 10.500/15.000              ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3, "unit": "Kg"},
                None,
                [
                    "■■■■□ task - 10.500/15.000Kg            ",
                ],
            ),
            (
                (0.3,),
                {},
                "comment!",
                [
                    "■■□□□ task - 30.00% - comment!          ",
                ],
            ),
            (
                (0.3,),
                {"unit": "whatever"},
                "comment!",
                [
                    "■■□□□ task - 30.00% - comment!          ",
                ],
            ),
            (
                (0.3,),
                {"ndigits": 0},
                "comment!",
                [
                    "■■□□□ task - 30% - comment!             ",
                ],
            ),
            (
                (11, 15),
                {},
                "comment!",
                [
                    "■■■■□ task - 11/15 - comment!           ",
                ],
            ),
            (
                (10.5, 15),
                {},
                "comment!",
                [
                    "■■■■□ task - 10.50/15.00 - comment!     ",
                ],
            ),
            (
                (10.5, 15),
                {"unit": "Kg"},
                "comment!",
                [
                    "■■■■□ task - 10.50/15.00Kg - comment!   ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3},
                "comment!",
                [
                    "■■■■□ task - 10.500/15.000 - comment!   ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3, "unit": "Kg"},
                "comment!",
                [
                    "■■■■□ task - 10.500/15.000Kg - comment! ",
                ],
            ),
        ],
    )
    def test_progress(self, io_mocker: IOMocker, args, kwargs, comment, expected):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(expected)
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress(*args, **kwargs)
                if comment:
                    task.comment(comment)
                io_mocker.mark()

    @pytest.mark.parametrize(
        "args,kwargs,expected",
        [
            (
                (50, 100),
                {},
                [
                    "■■□□□ task - 50.00/100.00B              ",
                ],
            ),
            (
                (0.50, 1),
                {},
                [
                    "■■□□□ task - 0.50/1.00B                 ",
                ],
            ),
            (
                (50, 100 * 1024),
                {},
                [
                    "□□□□□ task - 50.00B/100.00K             ",
                ],
            ),
            (
                (50, 100 * 1024 + 100),
                {},
                [
                    "□□□□□ task - 50.00B/100.10K             ",
                ],
            ),
            (
                (50 * 1024, 100 * 1024),
                {},
                [
                    "■■□□□ task - 50.00/100.00K              ",
                ],
            ),
            (
                (50 * 1024, 100.1 * 1024),
                {},
                [
                    "■■□□□ task - 50.00/100.10K              ",
                ],
            ),
            (
                (50, 100.501 * 1024 * 1024),
                {},
                [
                    "□□□□□ task - 50.00B/100.50M             ",
                ],
            ),
            (
                (50, 100.501 * 1024 * 1024),
                {"ndigits": 4},
                [
                    "□□□□□ task - 50.0000B/100.5010M         ",
                ],
            ),
            (
                (1 * 1024 * 1024 * 1024, 100.5 * 1024 * 1024 * 1024),
                {},
                [
                    "□□□□□ task - 1.00/100.50G               ",
                ],
            ),
            (
                (1 * 1024 * 1024 * 1024, 5000 * 1024 * 1024 * 1024 * 1024 * 1024),
                {},
                [
                    "□□□□□ task - 1.00G/5000.00P             ",
                ],
            ),
        ],
    )
    def test_progress_size(self, io_mocker: IOMocker, args, kwargs, expected):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(expected)
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress_size(*args, **kwargs)
                io_mocker.mark()

    @pytest.mark.parametrize(
        "args,kwargs,expected",
        [
            (
                (50, 100),
                {},
                [
                    "■■□□□ task - 50.00/100.00               ",
                ],
            ),
            (
                (50000, 100000),
                {},
                [
                    "■■□□□ task - 50.00K/100.00K             ",
                ],
            ),
            (
                (0.01052, 0.1),
                {},
                [
                    "■□□□□ task - 10.52m/100.00m             ",
                ],
            ),
            (
                (0.000000000000000000000001, 1000000000000000000000000),
                {},
                [
                    "□□□□□ task - 1.00y/1.00Y                ",
                ],
            ),
            (
                (0.0000000000000000000000001, 1000200000000000000000000000),
                {},
                [
                    "□□□□□ task - 0.10y/1000.20Y             ",
                ],
            ),
            (
                (50, 100),
                {"unit": "V"},
                [
                    "■■□□□ task - 50.00V/100.00V             ",
                ],
            ),
            (
                (0.00050, 0.001001),
                {"unit": "V"},
                [
                    "■■□□□ task - 500.00µV/1.00mV            ",
                ],
            ),
            (
                (0.00050, 0.001001),
                {"ndigits": 4},
                [
                    "■■□□□ task - 500.0000µ/1.0010m          ",
                ],
            ),
        ],
    )
    def test_progress_scale(self, io_mocker: IOMocker, args, kwargs, expected):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(expected)
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress_scale(*args, **kwargs)
                io_mocker.mark()

    def test_reset_progress(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ task - 2/5                        ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress(2, 5)
                io_mocker.mark()
                task.progress(None)
                io_mocker.mark()

    @pytest.mark.parametrize(
        "kwargs,expected_format",
        [
            ({}, "{}/{}"),
            ({"ndigits": 2}, "{}.00/{}.00"),
            ({"unit": "x"}, "{}/{}x"),
            ({"unit": "y", "ndigits": 1}, "{}.0/{}.0y"),
        ],
    )
    def test_iter(self, io_mocker: IOMocker, kwargs, expected_format: str):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"□□□□□ task - {expected_format.format(0, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■□□□□ task - {expected_format.format(1, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■□□□ task - {expected_format.format(2, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■■□□ task - {expected_format.format(3, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■■■□ task - {expected_format.format(4, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■■■■ task - {expected_format.format(5, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"> task - done                           ",
            ]
        )

        elems = [1, 2, "A", "B", object()]

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                for i, elem in enumerate(task.iter(elems, **kwargs)):
                    assert elem is elems[i]
                    io_mocker.mark()
                io_mocker.mark()

    def test_multiple_tasks(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ T1                                    ",
                "⣿ T2                                    ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ T1 - 2/5                          ",
                "■■■■□ T2 - 4/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "■■□□□ T1 - 2/5                          ",
                "■■■■□ T2 - 4/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "■■□□□ T1 - 2/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "bar                                     ",
                "■■□□□ T1 - 2/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "bar                                     ",
                "> T1 - done                             ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "bar                                     ",
                "> T1 - done                             ",
                "baz                                     ",
            ]
        )

        with io_mocker.mock():
            t1 = yuio.io.Task("T1")
            t2 = yuio.io.Task("T2")
            io_mocker.mark()
            t1.progress(2, 5)
            t2.progress(4, 5)
            io_mocker.mark()
            yuio.io.info("foo")
            io_mocker.mark()
            t2.done()
            io_mocker.mark()
            yuio.io.info("bar")
            io_mocker.mark()
            t1.done()
            io_mocker.mark()
            yuio.io.info("baz")

    def test_subtask(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ T1                                    ",
                "  ⣿ T1.1                                ",
                "  ⣿ T1.2                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ T1                                    ",
                "  ⣿ T1.1                                ",
                "  ⣿ T1.2 - done                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "⣿ T1                                    ",
                "  ⣿ T1.1                                ",
                "  ⣿ T1.2 - done                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "⣿ T1                                    ",
                "  ■■□□□ T1.1 - 2/5                      ",
                "  ⣿ T1.2 - done                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T1 - done                             ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T1 - done                             ",
                "bar                                     ",
            ]
        )

        with io_mocker.mock():
            t1 = yuio.io.Task("T1")
            t11 = t1.subtask("T1.1")
            t12 = t1.subtask("T1.2")
            io_mocker.mark()
            t12.done()
            io_mocker.mark()
            yuio.io.info("foo")
            io_mocker.mark()
            t11.progress(2, 5)
            io_mocker.mark()
            t1.done()
            io_mocker.mark()
            yuio.io.info("bar")

    def test_progressbar_decoration(self, theme: yuio.theme.Theme, io_mocker: IOMocker):
        theme.progress_bar_start_symbol = "["
        theme.progress_bar_end_symbol = "]"
        theme.progress_bar_done_symbol = ">"
        theme.progress_bar_pending_symbol = "."

        io_mocker.expect_screen(
            [
                "[>>...] Task - 2/5                      ",
            ]
        )
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("Task") as task:
                task.progress(2, 5)
                io_mocker.mark()


class TestSuspendOutput:
    @pytest.fixture()
    def wrap_streams(self) -> bool:
        return True

    @pytest.mark.parametrize(
        "meth,args,expected",
        [
            (
                "info",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "warning",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "success",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error_with_tb",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "heading",
                ("Bar.",),
                [
                    "                    ",
                    "⣿ Bar.              ",
                    "                    ",
                ],
            ),
            (
                "md",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "br",
                (),
                [
                    "                    ",
                ],
            ),
            (
                "raw",
                (yuio.term.ColorizedString("Bar."),),
                [
                    "Bar.                ",
                ],
            ),
        ],
    )
    def test_simple(self, io_mocker: IOMocker, meth, args, expected):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                *expected,
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            with yuio.io.SuspendOutput():
                io_mocker.mark()
                getattr(yuio.io, meth)(*args)
                io_mocker.mark()

    @pytest.mark.parametrize(
        "meth",
        [
            lambda s: sys.stderr.write(s + "\n"),
            lambda s: sys.stdout.write(s + "\n"),
            lambda s: print(s),
        ],
    )
    def test_streams(self, io_mocker: IOMocker, meth):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "Bar.                ",
            ]
        )

        with io_mocker.mock(wrap_streams=True):
            yuio.io.info("Foo.")
            with yuio.io.SuspendOutput():
                io_mocker.mark()
                meth("Bar.")
                io_mocker.mark()

    @pytest.mark.parametrize(
        "meth,args,expected",
        [
            (
                "warning",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "success",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error_with_tb",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "heading",
                ("Bar.",),
                [
                    "                    ",
                    "⣿ Bar.              ",
                    "                    ",
                ],
            ),
            (
                "md",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "br",
                (),
                [
                    "                    ",
                ],
            ),
            (
                "raw",
                (yuio.term.ColorizedString("Bar.\n"),),
                [
                    "Bar.                ",
                ],
            ),
        ],
    )
    def test_ignore(self, io_mocker: IOMocker, meth, args, expected):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                *expected,
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                *expected,
                *expected,
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            with yuio.io.SuspendOutput() as o:
                io_mocker.mark()
                getattr(yuio.io, meth)(*args)
                getattr(o, meth)(*args)
                io_mocker.mark()

    def test_task(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "> Task - done       ",
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            io_mocker.mark()
            with yuio.io.SuspendOutput():
                task = yuio.io.Task("Task")
                io_mocker.mark()
                task.done()
                io_mocker.mark()

    def test_task_start_before_suspended(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "⣿ Task              ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "⣿ Task              ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "> Task - done       ",
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            task = yuio.io.Task("Task")
            io_mocker.mark()
            with yuio.io.SuspendOutput():
                io_mocker.mark()
            io_mocker.mark()
            task.done()

    def test_task_finish_while_suspended(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "⣿ Task              ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "> Task - done       ",
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            task = yuio.io.Task("Task")
            io_mocker.mark()
            with yuio.io.SuspendOutput():
                io_mocker.mark()
                task.done()
                io_mocker.mark()


class TestLog:
    @pytest.fixture(autouse=True)
    def setup_logger(self):
        import logging

        self.logger = logging.getLogger("yuio.test.test_log")
        self.logger.setLevel(logging.DEBUG)
        handler = yuio.io.Handler()
        self.logger.addHandler(handler)

        yield

        self.logger.removeHandler(handler)

    def test_log(self, ostream: io.StringIO):
        self.logger.debug("debug")
        self.logger.info("info")
        self.logger.warning("warning")
        self.logger.error("error")
        self.logger.critical("critical")

        value = ostream.getvalue()
        assert "yuio.test.test_log DEBUG debug" in value
        assert "yuio.test.test_log INFO info" in value
        assert "yuio.test.test_log WARNING warning" in value
        assert "yuio.test.test_log ERROR error" in value
        assert "yuio.test.test_log CRITICAL critical" in value

    def test_log_tb(self, ostream: io.StringIO):
        self.logger.debug("debug")
        try:
            raise RuntimeError("oh no!")
        except RuntimeError:
            self.logger.exception("something went wrong")

        value = ostream.getvalue()
        assert "something went wrong" in value
        assert "oh no!" in value

    def test_log_stack(self, ostream: io.StringIO):
        self.logger.info("hi there!", stack_info=True)

        value = ostream.getvalue()
        assert "hi there!" in value
        assert 'self.logger.info("hi there!", stack_info=True)' in value



import datetime
import os
import pathlib
import sys

os.environ["__YUIO_SPHINX_BUILD"] = "1"
sys.path.append(str(pathlib.Path(__file__).parent / "_code"))

import yuio

# -- Project information -----------------------------------------------------

project = "Yuio"
copyright = f"{datetime.date.today().year}, Tamika Nomara"
author = "Tamika Nomara"
release = version = yuio.__version__


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "yuio.ext.sphinx",
    "sphinx_design",
    "sphinx_vhs",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
nitpick_ignore_regex = [
    (r"py:class", r"(.*\.)?([A-Z]{1,2}|[A-Z]+_co|Cmp|SupportsLt|Sz|TAst|_[^.]*)")
]
autodoc_typehints_format = "short"
autodoc_member_order = "bysource"
autodoc_inherit_docstrings = False
autodoc_type_aliases = {
    "ActionKey": "yuio.widget.ActionKey",
    "ActionKeys": "yuio.widget.ActionKeys",
    "Action": "yuio.widget.Action",
    # "RawString": "yuio.term.RawString",
    # "AnyString": "yuio.term.AnyString",
    # "JsonValue": "yuio.json_schema.JsonValue",
    # "Disabled": "yuio.DISABLED",
    # "Missing": "yuio.MISSING",
    # "Positional": "yuio.POSITIONAL",
    # "Omit": "yuio.OMIT",
}

vhs_cwd = pathlib.Path(__file__).parent.parent.parent
vhs_min_version = "0.7.2"

# -- Options for HTML output -------------------------------------------------

html_theme = "furo"
html_static_path = ["_static"]
html_css_files = ["extra.css"]
html_js_files = [
    "mermaid-init.js",
    (
        "https://cdn.jsdelivr.net/npm/mermaid@11.4.1/dist/mermaid.min.js",
        {"defer": "defer"},
    ),
]
html_theme_options = {
    "source_repository": "https://github.com/taminomara/yuio",
    "source_branch": "main",
    "source_directory": "docs/source",
}



import yuio.io
import yuio.widget

if __name__ == "__main__":
    term = yuio.io.get_term()
    theme = yuio.io.get_theme()

    options = [
        yuio.widget.Option(
            display_text="a coloring book", value="I'll bring you five :3"
        ),
        yuio.widget.Option(display_text="a comic", value="Which series do you like?"),
        yuio.widget.Option(
            display_text="release of Half-Life 3", value="Maybe a dragon?"
        ),
        yuio.widget.Option(
            display_text="a dragon!", value="What color do you want your dragon to be?"
        ),
        yuio.widget.Option(
            display_text="gender euphoria 👉👈",
            value="Aww :3 you're loved and valid, remember that 💖",
            #     " (and I really hope you survive this shitty world!)"
        ),
    ]

    widget = yuio.widget.Choice(options)

    yuio.io.heading("What should Santa bring you this year?")
    result = widget.run(term, theme)
    yuio.io.success("%s", result)



from __future__ import annotations

import importlib
import pathlib
import types

import pytest

from yuio import _typing as _t

root = pathlib.Path(__file__).parent.parent / "yuio"
files = [file for file in root.glob("*.py") if not file.name.startswith("_")]


@pytest.mark.parametrize("file", files, ids=[file.name for file in files])
def test_export_consistency(file: pathlib.Path):
    path = "yuio." + ".".join(file.relative_to(root).parts).removesuffix(".py")
    module = importlib.import_module(path)
    all = set(module.__all__)
    missing = set()
    for name, value in module.__dict__.items():
        if name.startswith("_"):
            continue
        if name in all:
            continue
        if isinstance(value, (types.ModuleType, _t.TypeVar, _t.ParamSpec)):
            continue
        if obj_module := getattr(value, "__module__", None):
            if not obj_module.startswith("yuio"):
                continue
        missing.add(name)

    assert not missing, "some items are missing from __all__"



import yuio.app
import yuio.io

@yuio.app.app
def main():
    yuio.io.heading("Message colors")

    yuio.io.success("Success message is bold green.")
    yuio.io.failure("Failure message is bold red.")
    yuio.io.info("Info message is default color.")
    yuio.io.warning("Warning message is yellow.")
    yuio.io.error("Error message is red.")

if __name__ == "__main__":
    main.run()



import enum
import yuio.app
import yuio.io

class GreetingType(enum.Enum):
    FORMAL = "Formal"
    INFORMAL = "Informal"

@yuio.app.app
def main():
    name = yuio.io.ask("What's your name?")
    greeting_type = yuio.io.ask[GreetingType](
        "What kind of greeting do you want?",
        default=GreetingType.FORMAL,
    )

    if greeting_type is GreetingType.FORMAL:
        yuio.io.info("Hello, %s. I hope this email finds you well.", name)
    else:
        yuio.io.info("Hii %s! So good to see you!", name)

if __name__ == "__main__":
    main.run()



import time
import subprocess
import yuio.app
import yuio.io

@yuio.app.app
def main():
    with yuio.io.Task("Performing some task"):
        time.sleep(1)

        # All progress bars, prints, and so on are suspended
        # while `git status` runs.
        with yuio.io.SuspendOutput() as o:
            # But you can manually bypass output suspension.
            o.info("Running `git status`:")

            subprocess.check_call(["git", "status"])

        time.sleep(1)

if __name__ == "__main__":
    main.run()



import yuio.io
import yuio.widget

if __name__ == "__main__":
    term = yuio.io.get_term()
    theme = yuio.io.get_theme()

    widget = yuio.widget.Input(
        placeholder="Enter something nice?",
        decoration=">",
        allow_multiline=True,
        allow_special_characters=True,
    )

    result = widget.run(term, theme)
    yuio.io.success("You've entered `%r`", result)
